# -*- coding: utf-8 -*-

# -------------------------------------------#
# author: sean lee                           #
# email: xmlee97@gmail.com                   #
# -------------------------------------------#

__author__ = 'Sean Lee <xmlee97@gmail.com>'
__version__ = '0.5.2'


from typing import List, Tuple, Optional

from xmnlp import config
from xmnlp import lexical
from xmnlp import summary
from xmnlp import checker as _checker
from xmnlp import sentiment as _sentiment
from xmnlp import pinyin as _pinyin
from xmnlp import radical as _radical
from xmnlp.utils import load_stopword


seg = lexical.seg
tag = lexical.tag
fast_seg = lexical.fast_seg
fast_tag = lexical.fast_tag
deep_seg = lexical.deep_seg
deep_tag = lexical.deep_tag
ner = lexical.ner

pinyin = _pinyin.translate
radical = _radical.radical
checker = _checker.spellcheck
sentiment = _sentiment.sentiment


def set_model(dirname: str):
    config.MODEL_DIR = dirname


def set_stopword(fpath: str):
    """set stopwords from file"""
    config.SYS_STOPWORDS = set(load_stopword(fpath)) | config.SYS_STOPWORDS


def keyword(text: str,
            k: int = 10,
            stopword: bool = True,
            allowPOS: Optional[List[str]] = None) -> List[Tuple[str, float]]:
    """extract keyword from text
    Args:
      text: str
      k: int, 返回 topk 个关键词
      stopword: bool, 是否设置停用词
      allowPOS: Optional[List[str]], 允许的词性，默认为系统自定义的词性列表
    Return:
      List[Tuple[str, float]]
    """
    if allowPOS is None:
        allowPOS = config.ALLOW_POS
    stopwords = config.SYS_STOPWORDS if stopword else []

    return summary.keyword(text, k=k, stopword=stopwords, allowPOS=allowPOS)


def keyphrase(text: str,
              k: int = 10,
              stopword: bool = False) -> List[str]:
    """keyphrase extraction
    Args:
      text: str
      k: int, 返回 topk 个关键句
      stopword: bool, 是否设置停用词
      allowPOS: Optional[List[str]], 允许的词性, 默认为系统自定义的词性列表
    Return:
      List[str]
    """
    stopwords = config.SYS_STOPWORDS if stopword else []
    return summary.keyphrase(text, k=k, stopword=stopwords)



# -*- coding: utf-8 -*-

import os
import re
from typing import Optional, List, Tuple

from xmnlp.config.path import ROOT
from xmnlp.lexical import deep_tag


re_english = re.compile(r'[a-zA-Z0-9]+')


class Tokenization:
    def __init__(self,
                 user_vocab_path: Optional[str] = None,
                 detect_new_word: bool = True):
        """ Tokenization
        Arguments:
          user_vocab_path: str, 用户自定义字典, 默认 None
          detect_new_word: bool, 是否识别新词, 默认 True
        """
        # dict.big.txt 是系统词库
        vocab_paths = [os.path.join(ROOT, 'dict.big.txt')]
        if user_vocab_path is not None:
            vocab_paths.append(user_vocab_path)
        self.word2tag, self.max_word_length = self.load_vocab(vocab_paths)
        self.new_word_kernel = deep_tag if detect_new_word else None

    def seg(self, doc: str) -> List[str]:
        """ seg document
        Args:
          doc: str, input document
        """
        return [w[0] for w in self.tag(doc)]

    def tag(self, doc: str) -> List[Tuple[str, str]]:
        """ tag doc
        Args:
          doc: str, input doc
        """
        words = []
        word_length = len(doc)
        while word_length > 0:
            N = min(self.max_word_length, word_length)
            word = doc[-N:]
            while N > 0:
                if word in self.word2tag or N == 1:
                    words.append(word)
                    break
                N -= 1
                word = word[-N:]
            doc = doc[:-N]
            word_length -= N

        # 合并英文和数字
        combine_words = []
        tmp = ''
        for w in words[::-1]:
            if w.encode('utf-8').isalpha() or w.isdigit():
                tmp += w
            else:
                if tmp:
                    combine_words.append((tmp, 'eng'))
                if w.strip():
                    combine_words.append((w, self.word2tag.get(w, 'x')))
                tmp = ''
        if tmp:
            combine_words.append((tmp, 'eng'))
        words = combine_words

        if self.new_word_kernel is None:
            return words

        # new words detect
        final_words = []
        N = len(words)
        i, j = 0, 0
        while i < N:
            if len(words[i][0]) > 1:
                final_words.append(words[i])
                i += 1
                continue
            for k in range(i + 1, N):
                j = k
                if len(words[k][0]) > 1:
                    break
            if i + 1 == j:
                final_words.append(words[i])
            elif i + 1 == N:
                final_words.append(words[i])
                break
            else:
                sequence = ''.join([t[0] for t in words[i:j]])
                final_words += self.new_word_kernel(sequence)
            i = j
        return final_words

    def load_vocab(self, fpaths: List[str]) -> dict:
        word2tag = {}
        max_word_length = 0
        for fpath in fpaths:
            with open(fpath, 'r', encoding='utf-8') as reader:
                for line in reader:
                    line = line.strip()
                    if not line:
                        continue
                    arr = line.split()
                    assert len(arr) in [2, 3], 'supported vocabulary formats: 1)word tag 2)word frequency tag.'
                    if len(arr) == 2:
                        word, tag = arr
                    else:
                        word, _, tag = arr
                    word2tag[word] = tag
                    max_word_length = max(max_word_length, len(word))
        return word2tag, max_word_length



# -*- coding: utf-8 -*-

# -------------------------------------------#
# author: sean lee                           #
# email: xmlee97@gmail.com                   #
# -------------------------------------------#

""" XMNLP - Lexical

Model Tree:

lexical
├── label2id.json
├── lexical.onnx
└── vocab.txt
"""

import os
import re
import json
import logging
from typing import List, Tuple, Optional, Union

import numpy as np
from tokenizers import BertWordPieceTokenizer

from xmnlp.base_model import BaseModel
from xmnlp.utils import rematch


MAX_LEN = 512
re_split = re.compile(r'.*?[\n。]+')


class LexicalModel(BaseModel):
    def predict(self, token_ids: np.ndarray, segment_ids: np.ndarray) -> np.ndarray:
        token_ids = token_ids.astype('float32')
        segment_ids = segment_ids.astype('float32')
        return self.sess.run(['crf/cond_1/Merge'], {
            'Input-Token': token_ids,
            'Input-Segment': segment_ids
        })


class Lexical:
    def __init__(self, model_dir: Optional[str] = None):
        self.tokenizer = BertWordPieceTokenizer(os.path.join(model_dir, 'vocab.txt'), lowercase=True)
        self.tokenizer.enable_truncation(max_length=MAX_LEN)
        self.lexical_model = LexicalModel(os.path.join(model_dir, 'lexical.onnx'))
        with open(os.path.join(model_dir, 'label2id.json'), encoding='utf-8') as reader:
            label2id = json.load(reader)
        self.id2label = {int(v): k for k, v in label2id.items()}

    def predict_one(self, text: str, base_position: int = 0) -> List[Tuple[str, str, int, int]]:
        tokenized = self.tokenizer.encode(text)
        token_ids = tokenized.ids
        segment_ids = tokenized.type_ids
        mapping = rematch(tokenized.offsets)
        token_ids, segment_ids = np.array([token_ids]), np.array([segment_ids])
        logits = self.lexical_model.predict(token_ids, segment_ids)[0][0]
        labels = [self.id2label[i] for i in np.argmax(logits, axis=1)]
        res = []
        for s, e, t in self.bio_decode(labels):
            s = mapping[s]
            s = 0 if not s else s[0]
            e = mapping[e]
            e = len(text) - 1 if not e else e[-1]
            res.append((text[s: e + 1], t, base_position + s, base_position + e + 1))

        return res

    def bio_decode(self, labels: List[str]) -> List[Tuple[int, int, str]]:
        entities = []
        start_tag = None
        for i, tag in enumerate(labels):
            tag_capital = tag.split('-')[0]
            tag_name = tag.split('-')[1] if tag != 'O' else ''
            if tag_capital in ['B', 'O']:
                if start_tag is not None:
                    entities.append((start_tag[0], i - 1, start_tag[1]))
                    start_tag = None
                if tag_capital == 'B':
                    start_tag = (i, tag_name)
            elif tag_capital == 'I' and start_tag is not None and start_tag[1] != tag_name:
                entities.append((start_tag[0], i - 1, start_tag[1]))
                start_tag = (i, tag_name)
        if start_tag is not None:
            entities.append((start_tag[0], i, start_tag[1]))

        return entities

    def transform(self, data: List[Tuple[str, str, int, int]], with_position: bool) -> List[
        Union[Tuple[str, str], Tuple[str, str, int, int]]
    ]:
        if with_position:
            return data
        return [(w, t) for w, t, _, _ in data]

    def predict(self, text: str, with_position: bool = False) -> List[
        Union[Tuple[str, str], Tuple[str, str, int, int]]
    ]:
        if len(text) < MAX_LEN:
            return self.transform(self.predict_one(text), with_position)

        logging.warn('xmnlp: 处理的文本过长（>512），可能会得到意料之外的结果')
        sentences = re_split.findall(text)
        if not sentences:
            return self.transform(self.predict_one(text), with_position)

        result = []
        prev, base_position = 0, 0
        for i in range(1, len(sentences)):
            current_text = ''.join(sentences[prev: i])
            next_text = ''.join(sentences[prev: i + 1])
            if len(current_text) <= MAX_LEN and len(next_text) > MAX_LEN:
                result += self.predict_one(current_text, base_position=base_position)
                prev = i
                base_position += len(current_text)
        result += self.predict_one(''.join(sentences[prev:]), base_position=base_position)
        return self.transform(result, with_position)
